import torchvision
import os

transform = torchvision.transforms.Compose([torchvision.transforms.ToTensor(), torchvision.transforms.Normalize((0.2859,), (0.3530,))])
root = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'RAW_DATA', 'FASHION')
train_data = torchvision.datasets.FashionMNIST(root=root, download=True, train=True, transform=transform)
test_data = torchvision.datasets.FashionMNIST(root=root, download=True, train=False, transform=transform)


import torchvision
from flgo.benchmark.toolkits.cv.classification import BuiltinClassGenerator, BuiltinClassPipe, GeneralCalculator
import flgo.benchmark
import os.path

builtin_class = torchvision.datasets.EMNIST
transform = torchvision.transforms.Compose([torchvision.transforms.ToTensor()])
path = os.path.join(flgo.benchmark.data_root, 'EMNIST')

class TaskGenerator(BuiltinClassGenerator):
    def __init__(self, rawdata_path=path, split='byclass'):
        super(TaskGenerator, self).__init__(os.path.split(os.path.dirname(__file__))[-1], rawdata_path, builtin_class, transform)
        self.split = split
        self.additional_option = {'split': self.split}

    def load_data(self):
        self.train_data = torchvision.datasets.EMNIST(root=self.rawdata_path, split = self.split, download=True, train=True, transform=self.transform)
        self.test_data = torchvision.datasets.EMNIST(root=self.rawdata_path, split = self.split, download=True, train=False, transform=self.transform)

class TaskPipe(BuiltinClassPipe):
    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, builtin_class, transform)

TaskCalculator = GeneralCalculator


import copy
from .fedbase import BasicServer
from .fedbase import BasicClient
import flgo.utils.fmodule as fmodule
import torch

class Server(BasicServer):
    def initialize(self, *args, **kwargs):
        """
        lmbd: 0.001, 0.01, 0.05, 0.1
        alpha: 0.0, 0.5, 0.75, 0.875, 0.9375, 1.0
        rho: 0.1
        """
        self.init_algo_para({"lmbd": 0.01, 'alpha':0.9, 'rho':0.1})
        self.aggregation_option = 'uniform'
class Client(BasicClient):
    def initialize(self, *args, **kwargs):
        self.local_grad_controller = self.server.model.zeros_like().to('cpu')

    @fmodule.with_multi_gpus
    def train(self, model):
        model.train()
        global_model = copy.deepcopy(model)
        global_model.freeze_grad()
        self.local_grad_controller.to(self.device)
        his_sub_global = self.local_grad_controller - global_model
        optimizer = self.calculator.get_optimizer(model, lr=self.learning_rate, weight_decay=self.weight_decay+self.lmbd, momentum=self.momentum)
        for iter in range(self.num_steps):
            # get a batch of data
            optimizer.zero_grad()
            batch_data = self.get_batch_data()
            # Line 7 in Algo.1
            loss1 = self.calculator.compute_loss(model, batch_data)['loss']
            loss1.backward()
            grad1 = [p.grad for p in model.parameters()]
            # Line 8 in Algo.1
            grad1_norm = torch.norm(torch.cat([g.view(-1) for g in grad1 if g is not None]), 2)
            with torch.no_grad():
                for p, g in zip(model.parameters(), grad1):
                    p.data = p.data + self.rho * g / grad1_norm
            model.zero_grad()
            # Line 9 in Algo.1
            loss2 = self.calculator.compute_loss(model, batch_data)['loss']
            loss2.backward()
            grad2 = [p.grad for p in model.parameters()]
            model.zero_grad()
            with torch.no_grad():
                for p, g in zip(model.parameters(),grad1):
                    p.data = p.data - self.rho * g / grad1_norm
            # Line 10 in Algo.1
            quasi_grad = [(1.0-self.alpha)*g1+self.alpha*g2 if g1 is not None else None for g1, g2 in zip(grad1, grad2)]
            # Line 11 in Algo.1
            with torch.no_grad():
                for p, qg, hsgi in zip(model.parameters(), quasi_grad, his_sub_global.parameters()):
                    p.grad = qg - self.lmbd * hsgi
            if self.clip_grad > 0: torch.nn.utils.clip_grad_norm_(parameters=model.parameters(), max_norm=self.clip_grad)
            optimizer.step()
            # model_tmp.load_state_dict(copy.deepcopy(model.state_dict()))
        # Line 13 in Algo.1
        self.local_grad_controller = self.local_grad_controller + (model - global_model)
        # Line 14 in Algo.1
        model.load_state_dict((model + self.local_grad_controller).state_dict())
        self.local_grad_controller.to('cpu')
        return



import os
from flgo.benchmark.toolkits.cv.classification import GeneralCalculator, FromDatasetPipe, FromDatasetGenerator
from .config import train_data
try:
    from .config import test_data
except:
    test_data = None
try:
    from .config import val_data
except:
    val_data = None

class TaskGenerator(FromDatasetGenerator):
    def __init__(self):
        super(TaskGenerator, self).__init__(benchmark=os.path.split(os.path.dirname(__file__))[-1],
                                            train_data=train_data, val_data=val_data, test_data=test_data)

class TaskPipe(FromDatasetPipe):
    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, train_data=train_data, val_data=val_data, test_data=test_data)

TaskCalculator = GeneralCalculator


import os
import torchvision

transform = torchvision.transforms.Compose([torchvision.transforms.ToTensor(), torchvision.transforms.Normalize((0.485,0.456,0.406), (0.229,0.224,0.225))])
root = os.path.join(flgo.benchmark.data_root, 'SVHN')
train_data = torchvision.datasets.SVHN(root=root,transform=transform, download=True, split='train')
test_data = torchvision.datasets.SVHN(root=root, transform=transform, download=True, split='test')


import torchvision.transforms
from torch import nn
from flgo.utils.fmodule import FModule
from torch.utils.data import Dataset
from torchvision.transforms import RandomCrop, RandomHorizontalFlip

class Model(FModule):
    def __init__(self):
        super(Model, self).__init__()
        self.encoder = nn.Sequential(
            nn.Conv2d(3, 32, 3),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Conv2d(32, 64, 3),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Conv2d(64, 128, 3),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Flatten(1),
            nn.Linear(12800, 512),
            nn.ReLU(),
        )
        self.head = nn.Linear(512, 10)

    def forward(self, x):
        x = self.encoder(x)
        return self.head(x)

class AugmentDataset(Dataset):
    def __init__(self, dataset):
        self.dataset = dataset
        self.transform = torchvision.transforms.Compose([RandomCrop(size=(96, 96), padding=12), RandomHorizontalFlip(0.5)])

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, item):
        img, label = self.dataset[item]
        return self.transform(img), label

def init_dataset(object):
    if 'Client' in object.get_classname():
        object.train_data = AugmentDataset(object.train_data)

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


import flgo.algorithm.fedbase as fedbase
import torch
from tqdm import tqdm
import ray
import torch.utils.data as tud
import copy
import flgo.simulator.base as ss
import flgo.utils.shared_memory as fus
import warnings
warnings.filterwarnings('ignore')

class Server(fedbase.BasicServer):
    def initialize(self, *args, **kwargs):
        self.init_algo_para({'mu':0.1})
        self.sample_option = 'md' if self.proportion < 1.0 else 'full'
        self.aggregation_option = 'uniform'
    @ss.with_clock
    def communicate(self, selected_clients, mtype=0, asynchronous=False):
        """
        The whole simulating communication procedure with the selected clients.
        This part supports for simulating the client dropping out.

        Args:
            selected_clients (list of int): the clients to communicate with
            mtype (anytype): type of message
            asynchronous (bool): asynchronous communciation or synchronous communcation

        Returns:
            :the unpacked response from clients that is created ny self.unpack()
        """
        packages_received_from_clients = []
        received_package_buffer = {}
        communicate_clients = list(set(selected_clients))
        # prepare packages for clients
        for client_id in communicate_clients:
            received_package_buffer[client_id] = None
        # communicate with selected clients
        if self.num_parallels <= 1:
            # computing iteratively
            for client_id in tqdm(communicate_clients, desc="Local Training on {} Clients".format(len(communicate_clients)), leave=False):
                server_pkg = self.pack(client_id, mtype)
                server_pkg['__mtype__'] = mtype
                response_from_client_id = self.communicate_with(self.clients[client_id].id, package=server_pkg)
                packages_received_from_clients.append(response_from_client_id)
        else:
            paralleltype = self.option.get('parallel_type', 'obj')
            @ray.remote(num_gpus=torch.cuda.device_count()*1.0/self.num_parallels)
            def client_train_fedprox(data_ref, model, config, calculator):
                if config['parallel_type'] != 'obj':
                    data = fus.sharable2dataset(data_ref)
                else:
                    data = data_ref
                device = calculator.device
                data_loader = tud.DataLoader(data, batch_size=config['batch_size'], shuffle=True)
                optimizer = calculator.get_optimizer(model, lr=config['lr'], weight_decay=config['weight_decay'], momentum=config['momentum'])
                model.train()
                model.to(device)
                mu = config['mu']
                src_model = copy.deepcopy(model)
                src_model.freeze_grad()
                for epoch in range(config['num_epochs']):
                    for i, batch in enumerate(data_loader):
                        batch = calculator.to_device(batch)
                        model.zero_grad()
                        loss = calculator.compute_loss(model, batch)['loss']
                        loss_proximal = 0
                        for pm, ps in zip(model.parameters(), src_model.parameters()):
                            loss_proximal += torch.sum(torch.pow(pm - ps, 2))
                        loss = loss + 0.5 * mu * loss_proximal
                        loss.backward()
                        if config.get('clip_grad', -1) > 0: torch.nn.utils.clip_grad_norm_(parameters=model.parameters(), max_norm=config.get('clip_grad', 0.))
                        optimizer.step()
                return {'model': model.to('cpu')}
            self.model = self.model.to(torch.device('cpu'))
            if not hasattr(self, '_client_data_sharable'):
                if paralleltype != 'obj':
                    self._client_data_sharable = [fus.dataset2sharable(c.train_data) for c in self.clients]
                else:
                    self._client_data_sharable = [ray.put(c.train_data) for c in self.clients]
            for client_id in communicate_clients:
                c = self.clients[client_id]
                model = self.model
                config = {'mu': c.mu, 'lr':c.learning_rate, 'weight_decay':c.weight_decay, 'momentum':c.momentum, 'batch_size':c.batch_size, 'num_epochs':c.num_epochs, 'clip_grad':c.clip_grad}
                config['parallel_type'] = paralleltype
                calculator = self.gv.TaskCalculator(device='cuda',  optimizer_name=self.option.get('optimizer', 'SGD'),)
                cpkg = client_train_fedprox.remote(self._client_data_sharable[client_id], model, config, calculator)
                packages_received_from_clients.append(cpkg)
            ready_refs, remaining_refs = ray.wait(packages_received_from_clients, num_returns=len(communicate_clients), timeout=None)
            packages_received_from_clients = ray.get(ready_refs)
            self.model = self.model.to(self.device)
            for pkg in packages_received_from_clients:
                for k, v in pkg.items():
                    if hasattr(v, 'to'):
                        try:
                            pkg[k] = v.to(self.device)
                        except:
                            continue
        for i, client_id in enumerate(communicate_clients): received_package_buffer[client_id] = \
        packages_received_from_clients[i]
        packages_received_from_clients = [received_package_buffer[cid] for cid in selected_clients if
                                          received_package_buffer[cid]]
        self.received_clients = selected_clients
        return self.unpack(packages_received_from_clients)

class Client(fedbase.BasicClient):
    pass


r"""
This module is a template for benchmark of link prediction within a graph in horizontalFL. To use this module,
one can write codes as below

Example::

"""
from .model import default_model
import flgo.benchmark.toolkits.visualization
import flgo.benchmark.toolkits.partition

default_model = default_model
default_partitioner = flgo.benchmark.toolkits.partition.NodeLouvainPartitioner
default_partition_para = {'num_clients':10}
visualize = flgo.benchmark.toolkits.visualization.visualize_by_community




import torch
import flgo.benchmark.toolkits.cv.detection.transforms as T


class DetectionPresetTrain:
    def __init__(self, *, data_augmentation, hflip_prob=0.5, mean=(123.0, 117.0, 104.0)):
        if data_augmentation == "hflip":
            self.transforms = T.Compose(
                [
                    T.RandomHorizontalFlip(p=hflip_prob),
                    T.PILToTensor(),
                    T.ConvertImageDtype(torch.float),
                ]
            )
        elif data_augmentation == "lsj":
            self.transforms = T.Compose(
                [
                    T.ScaleJitter(target_size=(1024, 1024)),
                    T.FixedSizeCrop(size=(1024, 1024), fill=mean),
                    T.RandomHorizontalFlip(p=hflip_prob),
                    T.PILToTensor(),
                    T.ConvertImageDtype(torch.float),
                ]
            )
        elif data_augmentation == "multiscale":
            self.transforms = T.Compose(
                [
                    T.RandomShortestSize(
                        min_size=(480, 512, 544, 576, 608, 640, 672, 704, 736, 768, 800), max_size=1333
                    ),
                    T.RandomHorizontalFlip(p=hflip_prob),
                    T.PILToTensor(),
                    T.ConvertImageDtype(torch.float),
                ]
            )
        elif data_augmentation == "ssd":
            self.transforms = T.Compose(
                [
                    T.RandomPhotometricDistort(),
                    T.RandomZoomOut(fill=list(mean)),
                    T.RandomIoUCrop(),
                    T.RandomHorizontalFlip(p=hflip_prob),
                    T.PILToTensor(),
                    T.ConvertImageDtype(torch.float),
                ]
            )
        elif data_augmentation == "ssdlite":
            self.transforms = T.Compose(
                [
                    T.RandomIoUCrop(),
                    T.RandomHorizontalFlip(p=hflip_prob),
                    T.PILToTensor(),
                    T.ConvertImageDtype(torch.float),
                ]
            )
        else:
            raise ValueError(f'Unknown data augmentation policy "{data_augmentation}"')

    def __call__(self, img, target):
        return self.transforms(img, target)


class DetectionPresetEval:
    def __init__(self):
        self.transforms = T.Compose(
            [
                T.PILToTensor(),
                T.ConvertImageDtype(torch.float),
            ]
        )

    def __call__(self, img, target):
        return self.transforms(img, target)


import os
from .config import train_data # config必须包含train_data
try:
    from .config import test_data # config可选包含test_data和val_data
except:
    test_data = None
try:
    from .config import val_data
except:
    val_data = None
try:
    import ujson as json
except:
    import json
try:
    from .config import DataLoader as MyDataloader
except:
    MyDataloader = None
try:
    from .config import collate_fn
except:
    collate_fn = None
try:
    from .config import split_dataset
except:
    split_dataset = None
from .config import data_to_device, eval, compute_loss

import flgo.benchmark.base as fbb
import torch.utils.data
import os

class TaskGenerator(fbb.FromDatasetGenerator):
    def __init__(self):
        super(TaskGenerator, self).__init__(benchmark=os.path.split(os.path.dirname(__file__))[-1],
                                            train_data=train_data, val_data=val_data, test_data=test_data)

class TaskPipe(fbb.FromDatasetPipe):
    TaskDataset = torch.utils.data.Subset
    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, train_data=train_data, val_data=val_data, test_data=test_data)
        self.my_split_dataset = split_dataset

    def split_dataset(self, dataset, p=0.0):
        if dataset is None: return None, None
        if self.my_split_dataset is None:
            if p == 0: return dataset, None
            s1 = int(len(dataset) * p)
            s2 = len(dataset) - s1
            if s1==0:
                return dataset, None
            elif s2==0:
                return None, dataset
            else:
                return torch.utils.data.random_split(dataset, [s2, s1])
        else:
            return self.my_split_dataset(dataset, p)

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names}
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_datas[cid],}
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def load_data(self, running_time_option: dict) -> dict:
        train_data = self.train_data
        test_data = self.test_data
        val_data = self.val_data
        if val_data is None:
            server_data_test, server_data_val = self.split_dataset(test_data, running_time_option['test_holdout'])
        else:
            server_data_test, server_data_val = test_data, val_data
        task_data = {'server': {'test': server_data_test, 'val': server_data_val}}
        for cid, cname in enumerate(self.feddata['client_names']):
            cdata = self.TaskDataset(train_data, self.feddata[cname]['data'])
            cdata_train, cdata_val = self.split_dataset(cdata, running_time_option['train_holdout'])
            if running_time_option['train_holdout']>0 and running_time_option['local_test']:
                cdata_val, cdata_test = self.split_dataset(cdata_val, running_time_option['local_test_ratio'])
            else:
                cdata_test = None
            task_data[cname] = {'train':cdata_train, 'val':cdata_val, 'test': cdata_test}
        return task_data

class TaskCalculator(fbb.BasicTaskCalculator):
    r"""
    Support task-specific computation when optimizing models, such
    as putting data into device, computing loss, evaluating models,
    and creating the data loader
    """

    def __init__(self, device, optimizer_name='sgd'):
        super(TaskCalculator, self).__init__(device, optimizer_name)
        self.DataLoader = MyDataloader if MyDataloader is not None else torch.utils.data.DataLoader
        self.collate_fn = collate_fn

    def to_device(self, data, *args, **kwargs):
        return data_to_device(data, self.device)

    def get_dataloader(self, dataset, batch_size=64, *args, **kwargs):
        return self.DataLoader(dataset, batch_size=batch_size, collate_fn=self.collate_fn, **kwargs,)

    @torch.no_grad()
    def test(self, model, data, batch_size=64, num_workers=0, pin_memory=False, **kwargs):
        data_loader = self.get_dataloader(data, batch_size=64, num_workers=num_workers, pin_memory=pin_memory)
        return eval(model, data_loader, self.device)

    def compute_loss(self, model, data, *args, **kwargs):
        return compute_loss(model, data, self.device)


from flgo.benchmark.toolkits.tabular.classification import BuiltinClassGenerator, BuiltinClassPipe, GeneralCalculator
from flgo.benchmark.toolkits.tabular.classification.datasets import BankMarketing
import flgo.benchmark
import os.path

builtin_class = BankMarketing
path = os.path.join(flgo.benchmark.data_root,  'BANKMARKETING')

TaskCalculator = GeneralCalculator

class TaskGenerator(BuiltinClassGenerator):
    def __init__(self, rawdata_path=path):
        super(TaskGenerator, self).__init__(os.path.split(os.path.dirname(__file__))[-1], rawdata_path, builtin_class)

class TaskPipe(BuiltinClassPipe):
    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, builtin_class)


import torch_geometric
import os
import random
import torch.nn.functional as F
import torch_geometric.nn as pyg_nn
import torch.nn as nn
import torch
from torch.nn import Sequential, Linear, ReLU

path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'RAW_DATA', 'ENZYMES')
all_data = torch_geometric.datasets.TUDataset(path, name='ENZYMES', use_node_attr=True)
all_idxs = [i for i in range(len(all_data))]
random.seed(1)
random.shuffle(all_idxs)
num_samples = len(all_data)
train_idxs = all_idxs[:int(0.8*num_samples)]
test_idxs = all_idxs[int(0.8*num_samples):]
train_data = all_data[train_idxs]
test_data = all_data[test_idxs]

class GIN(torch.nn.Module):
    def __init__(self, input_dim=21, hidden_dim=64, output_dim=6):
        super(GIN, self).__init__()
        self.num_layers = 3
        self.pre_mp = nn.Sequential(
            nn.Linear(input_dim, hidden_dim))

        self.convs = nn.ModuleList()
        self.nn1 = Sequential(Linear(hidden_dim, hidden_dim), ReLU(), Linear(hidden_dim, hidden_dim))
        self.convs.append(pyg_nn.GINConv(self.nn1))
        for l in range(self.num_layers - 1):
            self.nnk = Sequential(Linear(hidden_dim, hidden_dim), ReLU(), Linear(hidden_dim, hidden_dim))
            self.convs.append(pyg_nn.GINConv(self.nnk))

        self.post_mp = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim), ReLU(), nn.Linear(hidden_dim, output_dim))

    def forward(self, data):
        x, edge_index, batch = data.x, data.edge_index, data.batch
        x = self.pre_mp(x)
        for i in range(len(self.convs)):
            x = self.convs[i](x, edge_index)
            x = F.relu(x)
        x = pyg_nn.global_add_pool(x, batch)
        x = self.post_mp(x)
        x = F.log_softmax(x, dim=1)
        return x

def get_model():
    return GIN()


try:
    import ujson as json
except:
    import json
import os
import uuid
import flgo.utils.fflow as fuf
import torch
import numpy as np
import torch.utils.data as tud
import sys
import pickle
from tqdm import tqdm

def get_dict_size(obj, seen=None):
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    seen.add(obj_id)
    size = sys.getsizeof(obj)
    if isinstance(obj, dict):
        for key, value in obj.items():
            size += sys.getsizeof(key) + get_dict_size(value, seen)
    elif isinstance(obj, (list, tuple, set)):
        for item in obj:
            size += get_dict_size(item, seen)
    return size

TYPE_CANDIDATES = ['int', 'float', 'str', 'ndarray', 'Tensor', 'list', 'tuple', 'dict', 'int64', 'float64']

class MemDataset(tud.Dataset):
    def __init__(self, data):
        super().__init__()
        self.data = [d for d in data]

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return len(self.data)

class TmpDataset(tud.Dataset):
    def __init__(self, data):
        super().__init__()
        self.data = data

    def __len__(self):
        return len(self.data[0])

    def __getitem__(self, i):
        return tuple(d[i] for d in self.data)

def _check_vector_shapes(vec_list):
    """
    Check whether the tensors\ndarrays have the same shape

    Args:
        vec_list (list[torch.Tensor]|list[numpy.ndarray]): list of vectors

    Returns:
        if_same (bool): True if the vectors have the same shape
    """
    if not vec_list: return True
    reference_shape = vec_list[0].shape
    for tensor in vec_list[1:]:
        if tensor.shape != reference_shape:
            return False
    return True

def dataset2sharable(dataset):
    """
    Convert a dataset into sharable format, i.e., numpy arrays of features and the type information for recovering them

    Args:
        dataset (torch.utils.data.Dataset): dataset to be converted

    Returns:
        sharable_data (list[numpy.ndarray]): the numpy arrays of the dataset
    """
    first_item = dataset[0]
    item_size = len(first_item)
    if not isinstance(first_item, tuple): first_item = tuple(first_item)
    etypes = [type(ei).__name__ if type(ei) not in TYPE_CANDIDATES else 'unknown' for ei in first_item]
    dataset_size = len(dataset)
    res = [[] for _ in range(item_size)]
    for i in range(dataset_size):
        for j in range(item_size):
            res[j].append(dataset[i][j])
    for j in range(item_size):
        if etypes[j] in ['int', 'float', 'str', 'int64', 'float64']:
            res[j] = np.array(res[j])
        elif etypes[j] == 'ndarray':
            if _check_vector_shapes(res[j]):
                res[j] = np.stack(res[j])
            else:
                shapes = [rjk.shape for rjk in res[j]]
                etypes[j] = {'etype': etypes[j], 'shape': shapes}
                res[j] = np.concatenate([rjk.reshape((-1, )) for rjk in res[j]])
        elif etypes[j] == 'Tensor':
            if _check_vector_shapes(res[j]):
                res[j] = torch.stack(res[j]).numpy()
            else:
                shapes = [tuple(rjk.shape) for rjk in res[j]]
                etypes[j] = {'etype': etypes[j], 'shape': shapes}
                res[j] = torch.cat([rjk.view(-1) for rjk in res[j]]).numpy()
        elif etypes[j] in ['list', 'tuple']:
            try:
                new_resj = [np.array(rjk) for rjk in res[j]]
                if _check_vector_shapes(new_resj):
                    new_resj = np.stack(new_resj)
                else:
                    shapes = [rjk.shape for rjk in new_resj]
                    etypes[j] = {'etype': etypes[j], 'shape': shapes}
                    new_resj = np.concatenate([rjk.reshape((-1, )) for rjk in new_resj])
                res[j] = new_resj
            except:
                res[j] = np.frombuffer(pickle.dumps(res[j]), dtype=np.uint8)
                etypes[j] = 'pickle'
        else:
            res[j] = np.frombuffer(pickle.dumps(res[j]), dtype=np.uint8)
            etypes[j] = 'pickle'
    etypes = np.frombuffer(pickle.dumps(etypes), dtype=np.uint8)
    res.append(etypes)
    return res

def sharable2dataset(sharable_data):
    """
    Recover the original data from sharable format

    Args:
        sharable_data (list[numpy.ndarray]): the numpy arrays of the dataset

    Returns:
        dataset (torch.utils.data.Dataset): the original dataset
    """

    types = pickle.loads(sharable_data.pop(-1).tobytes())
    data = []
    for i in range(len(types)):
        if isinstance(types[i], dict):
            etype = types[i]['etype']
            shapes = types[i]['shape']
            offsets = [0] + np.cumsum([np.prod(s) for s in shapes]).tolist()
            tmp_data = [sharable_data[i][offsets[k]:offsets[k + 1]].reshape(shapes[k]) for k in range(len(shapes))]
            if etype == 'Tensor':
                tmp_data = [torch.from_numpy(tdi) for tdi in tmp_data]
            elif etype == 'list':
                tmp_data = [tdi.tolist() for tdi in tmp_data]
            elif etype == 'tuple':
                tmp_data = [tuple(tdi.tolist()) for tdi in tmp_data]
            data.append(tmp_data)
        elif types[i] == 'pickle':
            data.append(pickle.loads(sharable_data[i].tobytes()))
        else:
            if types[i] == 'ndarray':
                data.append(sharable_data[i])
            elif types[i] == 'Tensor':
                data.append(torch.from_numpy(sharable_data[i]))
            elif types[i] in ['int', 'float', 'str']:
                data.append(sharable_data[i].tolist())
            elif types[i] in ['int64', 'float64']:
                data.append(sharable_data[i])
    return TmpDataset(data)

def create_memmap_meta_for_dataset(sharable_data, name, use_uuid=True):
    """
    Map sharable data to shared memory by np.memmap

    Args:
        sharable_data (list[numpy.ndarray]): the numpy arrays of the dataset
        name (str): the file name of the memmap
        use_uuid (bool): whether to use uuid for generating the file name

    Returns:
        shm_name (str): the file name of the memmap
        dtype (np.dtype): the data type of the memmap
    """
    dtype = np.dtype([(f'{i}', sdi.dtype, sdi.shape) for i, sdi in enumerate(sharable_data)])
    shm_name = name+f"{uuid.uuid4()}" if use_uuid else name
    shm = np.memmap(shm_name, dtype=dtype, mode='w+',shape=())
    for i in range(len(sharable_data)):
        shm[f'{i}'] = sharable_data[i]
        # np.copyto(shm[f'{i}'], sharable_data[i])
    return shm_name, dtype

def create_memmap_meta_for_task(task_data, path=''):
    """
    Map task data to shared memory by np.memmap

    Args:
        task_data (dict): the task data
        path (str): the path of the temporary dictionary to store the memmap files

    Returns:
        task_meta (dict): the meta information of the task memmap
    """
    task_meta = {}
    for party in task_data:
        task_meta[party] = {}
        for data_name in task_data[party]:
            data = task_data[party][data_name]
            if data is None: continue
            sharable_data = dataset2sharable(data)
            shm_name = "_".join([party, data_name])
            if path!='': shm_name = os.path.join(os.path.abspath(path), shm_name)
            shm_name, dtype = create_memmap_meta_for_dataset(sharable_data, shm_name)
            task_meta[party][data_name] = {
                "name": shm_name,
                "dtype": dtype,
            }
    return task_meta

def load_dataset_from_memmap_meta(name, dtype):
    """
    Load one dataset from np.memmap setting's value

    Args:
        name (str): shared_memory name
        dtype (list): list of dtypes
        etype (list(str)): the element type of items in original dataset
    Returns:
        party_data (torch.utils.data.Dataset): the recovered dataset
    """
    memmap = np.memmap(name, mode='r', dtype=dtype)
    sharable_data = [memmap[f'{i}'][0] for i in range(len(dtype))]
    return sharable2dataset(sharable_data)

def load_taskdata_from_memmap_meta(task_meta):
    """
    Load task data from np.memmap setting created by create_meta_for_task

    Args:
        task_meta (dict): task meta generated by create_meta_for_task

    Returns:
        task_data (dict): task data
    """
    task_data = {}
    for party in task_meta:
        task_data[party] = {}
        for data_name in task_meta[party]:
            task_data[party][data_name] = load_dataset_from_memmap_meta(**task_meta[party][data_name])
    return task_data

def create_task_data_npz(task, train_holdout:float=0.2, test_holdout:float=0.0, local_test=False, local_test_ratio=0.5, seed=0):
    """
    Convert task data into .npz files and store them at task/.mmap/task_config_info/*.npz

    Args:
        task (str): the path of the task
        train_holdout       (flt):  the rate of holding out the validation dataset from all the local training datasets',default=0.1
        test_holdout        (flt):  the rate of holding out the validation dataset from the testing datasets owned by the server', default=0.0
        local_test          (bool): if this term is set True and train_holdout>0, (0.5*train_holdout) of data will be set as client.test_data.default=False.
        local_test_ratio    (flt):  valid if local_test is True. The ratio of local test dataset holdout from local validation data.
        dataseed            (int):  seed for random initialization for data train/val/test partition', default=0

    Returns:
        if_success          (bool): True if the task data has been successfully converted
    """
    if not os.path.exists(task): return False
    memmap_path = os.path.join(task, '.mmap', "L{:.2f}G{:.2f}T{}R{:.2f}S{}".format(train_holdout, test_holdout, int(local_test), local_test_ratio, seed))
    if not os.path.exists(memmap_path): os.makedirs(memmap_path)
    else: raise FileExistsError(f'{memmap_path} already exists')
    task_data = fuf.load_task_data(task,  train_holdout, test_holdout, local_test, local_test_ratio, seed)
    task_meta = {}
    for party in tqdm(task_data, desc='Creating Task: '):
        task_meta[party] = {}
        for data_name in task_data[party]:
            data = task_data[party][data_name]
            if data is None: continue
            sharable_data = dataset2sharable(data)
            file_name = os.path.join(os.path.abspath(memmap_path), "_".join([party, data_name])+'.npz')
            saved_data = {f'{i}': sharable_data[i] for i in range(len(sharable_data))}
            np.savez(file_name, **saved_data)
            task_meta[party][data_name] = {
                "name": file_name,
            }
    meta_file = os.path.join(memmap_path, 'meta.json')
    with open(meta_file, 'w') as f:
        json.dump(task_meta, f)
    print("Task data has been successfully converted into .npz")
    return True

def load_task_data_from_npz(task, train_holdout:float=0.2, test_holdout:float=0.0, local_test=False, local_test_ratio=0.5, seed=0, create=False):
    """
    Load task data from task/.mmap/task_config_info/*.npz in memmap way

    Args:
        task (str): the path of the task
        train_holdout       (flt):  the rate of holding out the validation dataset from all the local training datasets',default=0.1
        test_holdout        (flt):  the rate of holding out the validation dataset from the testing datasets owned by the server', default=0.0
        local_test          (bool): if this term is set True and train_holdout>0, (0.5*train_holdout) of data will be set as client.test_data.default=False.
        local_test_ratio    (flt):  valid if local_test is True. The ratio of local test dataset holdout from local validation data.
        dataseed            (int):  seed for random initialization for data train/val/test partition', default=0
        create              (bool): create the npz file of the task if the target files doesn't exist

    Returns:
        task_data           (dict): the task data
    """
    if not os.path.exists(task):
        raise FileNotFoundError("Fedtask {} doesn't exist".format(task))
    memmap_path = os.path.join(task, '.mmap',"L{:.2f}G{:.2f}T{}R{:.2f}S{}".format(train_holdout, test_holdout, int(local_test), local_test_ratio, seed))
    if not os.path.exists(memmap_path):
        if create:
            create_task_data_npz(task, train_holdout, test_holdout, local_test, local_test_ratio, seed)
        else:
            raise FileNotFoundError("Task npy files were not found. Use flgo.utils.shared_memory.create_task_npy to create the target.")
    meta_file = os.path.join(memmap_path, 'meta.json')
    with open(meta_file, 'r') as f:
        task_meta = json.load(f)
    task_data = {}
    for party in tqdm(task_meta, desc='Loading Task: '):
        task_data[party] = {}
        for data_name in task_meta[party]:
            party_data = task_meta[party][data_name]
            file_name = party_data['name']
            sharable_data = np.load(file_name, mmap_mode='r')
            sharable_data = [sharable_data[k] for k in sharable_data.files]
            dataset = sharable2dataset(sharable_data)
            task_data[party][data_name] = dataset
    return task_data

if __name__=='__main__':
    def worker(task_meta):
        pid = os.getpid()
        task_data = load_taskdata_from_memmap_meta(task_meta)
        selected_party = 'server'
        dataset = task_data[selected_party]['test']
        dataloader = tud.DataLoader(dataset, batch_size=2)
        for batch_id, batch_data in enumerate(dataloader):
            x, y = batch_data
            print(f"Process {pid} Xid = {id(x.numpy())} {pid} Y = {id(y.numpy())}")
            if batch_id >= 3: break
        return True
    class TenDataset(tud.Dataset):
        def __init__(self, n: int = 5, l: int = 3, r: int = 8):
            self.n = n
            self.x = [torch.rand((torch.randint(l, r, (1,)).item())) for i in range(n)]
            self.y = torch.randint(0, 9, (n,)).long()

        def __len__(self):
            return self.n

        def __getitem__(self, i):
            return self.x[i], self.y[i]


    class NPDataset(tud.Dataset):
        def __init__(self, n: int = 5, l: int = 3, r: int = 8):
            self.n = n
            self.x = [torch.rand((torch.randint(l, r, (1,)).item())).numpy() for i in range(n)]
            self.y = torch.randint(0, 9, (n,)).long().numpy()

        def __len__(self):
            return self.n

        def __getitem__(self, i):
            return self.x[i], self.y[i]


    class DictData(tud.Dataset):
        def __init__(self, n: int = 5, l: int = 3, r: int = 8):
            self.n = n
            self.x = [{'x': torch.rand((torch.randint(l, r, (1,)).item())).numpy()} for i in range(n)]
            self.y = torch.randint(0, 9, (n,)).long().numpy()

        def __len__(self):
            return self.n

        def __getitem__(self, i):
            return self.x[i], self.y[i]
    NUM_WORKERS = 3
    # task = './my_task'
    DATA = NPDataset
    # task_data = fuf.load_task_data(task)
    task_data = {'server':{'test': DATA(), 'val':None}, 'Client01':{'train':DATA(), 'val':DATA(), 'test':None}}
    # DEBUG: create_meta_for_task
    task_meta = {}

    tmp_data =DATA()
    sharable_data = dataset2sharable(tmp_data)
    dataset = sharable2dataset(sharable_data)
    print('ok')
    # for party in task_data:
    #     task_meta[party] = {}
    #     for data_name in task_data[party]:
    #         data = task_data[party][data_name]
    #         if data is None: continue
    #         sharable_data, etypes = dataset2sharable(data)
    #         new_data = sharable2dataset(sharable_data, etypes)
    #         print(f"Consistency for {party} {data_name}: {all([all(new_data[i][0] == data[i][0]) for i in range(len(data))])}")
            # shm_name_list, shm_type_list, shm_shape_list= create_meta_for_dataset(sharable_data, "_".join([party, data_name]))
            # task_meta[party][data_name] = {
            #     "name": shm_name_list,
            #     "dtype": shm_type_list,
            #     "shape": shm_shape_list,
            #     'etype': etypes,
            # }
    # task_meta = create_meta_for_task(task_data)
    # worker(task_meta)
    # pros = []
    # for i in range(NUM_WORKERS):
    #     p = multiprocessing.Process(target=worker,  args=(task_meta, ))
    #     pros.append(p)
    # for p in pros: p.start()
    # for p in pros: p.join()



from flgo.utils import fmodule
import torchvision.models
import torch.nn as nn

class Model(fmodule.FModule):
    def __init__(self):
        super().__init__()
        resnet34 = torchvision.models.resnet34()
        resnet34.fc = nn.Linear(512, 10)
        resnet34.bn1 = nn.GroupNorm(num_groups=2, num_channels=64)

        resnet34.layer1[0].bn1 = nn.GroupNorm(num_groups=2, num_channels=64)
        resnet34.layer1[0].bn2 = nn.GroupNorm(num_groups=2, num_channels=64)
        resnet34.layer1[0].bn3 = nn.GroupNorm(num_groups=2, num_channels=64)
        resnet34.layer1[1].bn1 = nn.GroupNorm(num_groups=2, num_channels=64)
        resnet34.layer1[1].bn2 = nn.GroupNorm(num_groups=2, num_channels=64)
        resnet34.layer1[1].bn3 = nn.GroupNorm(num_groups=2, num_channels=64)

        resnet34.layer2[0].bn1 = nn.GroupNorm(num_groups=2, num_channels=128)
        resnet34.layer2[0].bn2 = nn.GroupNorm(num_groups=2, num_channels=128)
        resnet34.layer2[0].bn3 = nn.GroupNorm(num_groups=2, num_channels=128)
        resnet34.layer2[0].bn4 = nn.GroupNorm(num_groups=2, num_channels=128)
        resnet34.layer2[0].downsample[1] = nn.GroupNorm(num_groups=2, num_channels=128)
        resnet34.layer2[1].bn1 = nn.GroupNorm(num_groups=2, num_channels=128)
        resnet34.layer2[1].bn2 = nn.GroupNorm(num_groups=2, num_channels=128)
        resnet34.layer2[1].bn3 = nn.GroupNorm(num_groups=2, num_channels=128)
        resnet34.layer2[1].bn4 = nn.GroupNorm(num_groups=2, num_channels=128)

        resnet34.layer3[0].bn1 = nn.GroupNorm(num_groups=2, num_channels=256)
        resnet34.layer3[0].bn2 = nn.GroupNorm(num_groups=2, num_channels=256)
        resnet34.layer3[0].bn3 = nn.GroupNorm(num_groups=2, num_channels=256)
        resnet34.layer3[0].bn4 = nn.GroupNorm(num_groups=2, num_channels=256)
        resnet34.layer3[0].bn5 = nn.GroupNorm(num_groups=2, num_channels=256)
        resnet34.layer3[0].bn6 = nn.GroupNorm(num_groups=2, num_channels=256)
        resnet34.layer3[0].downsample[1] = nn.GroupNorm(num_groups=2, num_channels=256)
        resnet34.layer3[1].bn1 = nn.GroupNorm(num_groups=2, num_channels=256)
        resnet34.layer3[1].bn2 = nn.GroupNorm(num_groups=2, num_channels=256)
        resnet34.layer3[1].bn3 = nn.GroupNorm(num_groups=2, num_channels=256)
        resnet34.layer3[1].bn4 = nn.GroupNorm(num_groups=2, num_channels=256)
        resnet34.layer3[1].bn5 = nn.GroupNorm(num_groups=2, num_channels=256)
        resnet34.layer3[1].bn6 = nn.GroupNorm(num_groups=2, num_channels=256)


        resnet34.layer4[0].bn1 = nn.GroupNorm(num_groups=2, num_channels=512)
        resnet34.layer4[0].bn2 = nn.GroupNorm(num_groups=2, num_channels=512)
        resnet34.layer4[0].bn3 = nn.GroupNorm(num_groups=2, num_channels=512)
        resnet34.layer4[0].downsample[1] = nn.GroupNorm(num_groups=2, num_channels=512)
        resnet34.layer4[1].bn1 = nn.GroupNorm(num_groups=2, num_channels=512)
        resnet34.layer4[1].bn2 = nn.GroupNorm(num_groups=2, num_channels=512)
        resnet34.layer4[1].bn3 = nn.GroupNorm(num_groups=2, num_channels=512)
        self.model = resnet34

    def forward(self, x):
        return self.model(x)

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


import torch
from tqdm import tqdm
import torch.nn as nn
import os
import numpy as np
from flgo.benchmark.toolkits.cv.points_semantic_segmentation import GeneralCalculator, FromDatasetPipe, FromDatasetGenerator
from .config import train_data
try:
    from .config import test_data
except:
    test_data = None
try:
    from .config import val_data
except:
    val_data = None

def accuracy(scores, labels):
    r"""
        Compute the per-class accuracies and the overall accuracy #

        Parameters
        ----------
        scores: torch.FloatTensor, shape (B?, C, N)
            raw scores for each class
        labels: torch.LongTensor, shape (B?, N)
            ground truth labels

        Returns
        -------
        list of floats of length num_classes+1 (last item is overall accuracy)
    """
    num_classes = scores.size(-2) # we use -2 instead of 1 to enable arbitrary batch dimensions

    predictions = torch.max(scores, dim=-2).indices

    accuracies = []

    accuracy_mask = predictions == labels
    for label in range(num_classes):
        label_mask = labels == label
        per_class_accuracy = (accuracy_mask & label_mask).float().sum()
        per_class_accuracy /= label_mask.float().sum()
        accuracies.append(per_class_accuracy.cpu().item())
    # overall accuracy
    accuracies.append(accuracy_mask.float().mean().cpu().item())
    return accuracies

def intersection_over_union(scores, labels):
    r"""
        Compute the per-class IoU and the mean IoU #

        Parameters
        ----------
        scores: torch.FloatTensor, shape (B?, C, N)
            raw scores for each class
        labels: torch.LongTensor, shape (B?, N)
            ground truth labels

        Returns
        -------
        list of floats of length num_classes+1 (last item is mIoU)
    """
    num_classes = scores.size(-2) # we use -2 instead of 1 to enable arbitrary batch dimensions

    predictions = torch.max(scores, dim=-2).indices

    ious = []

    for label in range(num_classes):
        pred_mask = predictions == label
        labels_mask = labels == label
        iou = (pred_mask & labels_mask).float().sum() / (pred_mask | labels_mask).float().sum()
        ious.append(iou.cpu().item())
    ious.append(np.nanmean(ious))
    return ious

def evaluate(model, loader, criterion, device):
    model.eval()
    losses = []
    accuracies = []
    ious = []
    with torch.no_grad():
        for points, labels in tqdm(loader, desc='Validation', leave=False):
            points = points.to(device)
            labels = labels.to(device)
            scores = model(points)
            loss = criterion(scores, labels)
            losses.append(loss.cpu().item())
            accuracies.append(accuracy(scores, labels))
            ious.append(intersection_over_union(scores, labels))
    return {'loss':np.mean(losses),
            # 'accuracies':np.nanmean(np.array(accuracies), axis=0),
            # 'IoUs':np.nanmean(np.array(ious), axis=0),
            'mIoU': np.nanmean(np.array(ious), axis=0).mean(),
            'accuracy':np.nanmean(np.array(accuracies))
            }


class TaskGenerator(FromDatasetGenerator):
    def __init__(self):
        super(TaskGenerator, self).__init__(benchmark=os.path.split(os.path.dirname(__file__))[-1],
                                            train_data=train_data, val_data=val_data, test_data=test_data)

class TaskPipe(FromDatasetPipe):
    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, train_data=train_data, val_data=val_data, test_data=test_data)

class TaskCalculator(GeneralCalculator):
    def test(self, model, dataset, batch_size=64, num_workers=0, pin_memory=False):
        data_loader = self.DataLoader(dataset, batch_size=batch_size, num_workers=num_workers, pin_memory=pin_memory, collate_fn=self.collate_fn)
        return evaluate(model, data_loader, criterion=nn.CrossEntropyLoss(), device=self.device)


import torch
import torch.nn

EMBEDDING_SIZE = 64
NUM_CLASS = 10

class PartialMLPLayer(torch.nn.Module):
    def __init__(self, input_size, output_size):
        super().__init__()
        self.mlp = torch.nn.Linear(input_size, output_size)

    def forward(self, x):
        x = x.view(-1, x.shape[1] * x.shape[-2] * x.shape[-1])
        return self.mlp(x)

class Classifier(torch.nn.Module):
    def __init__(self, input_size, output_size):
        super().__init__()
        self.mlp = torch.nn.Linear(input_size, output_size)
        self.relu = torch.nn.ReLU()

    def forward(self, fusion):
        x = self.relu(fusion)
        x = self.mlp(x)
        return x

def init_local_module(object):
    partial_sample = object.train_data[0]
    feature = partial_sample[0]
    object.local_module = None if feature is None else PartialMLPLayer(len(feature.view(-1)), EMBEDDING_SIZE).to(object.device)

def init_global_module(object):
    partial_sample = object.train_data[0]
    feature, label, id = partial_sample
    object.global_module = None if label is None else Classifier(EMBEDDING_SIZE, NUM_CLASS).to(object.device)


from flgo.benchmark.base import *
from flgo.benchmark.toolkits.cv.classification import GeneralCalculator
from flgo.benchmark.toolkits.partition import BasicPartitioner
import numpy as np
TaskPipe = XYHorizontalTaskPipe
TaskCalculator = GeneralCalculator

class TaskGenerator(BasicTaskGenerator):
    def __init__(self, alpha=0.0, beta=0.0, num_clients=30, imbalance=0.0, mean_datavol=400, dimension=60, num_classes=10, *args, **kwargs):
        super().__init__('synthetic_regression', '')
        self.num_clients = num_clients
        self.alpha = alpha
        self.beta = beta
        self.imbalance = imbalance
        self.mean_datavol = mean_datavol
        self.dimension = dimension
        self.num_classes = num_classes

    def load_data(self):
        W_global = np.random.normal(0, 1, (self.dimension, self.num_classes))
        b_global = np.random.normal(0, 1, self.num_classes)
        v_global = np.zeros(self.dimension)
        # create Sigma = Diag([i^(-1.2) for i in range(60)])
        diagonal = np.zeros(self.dimension)
        for j in range(self.dimension):
            diagonal[j] = np.power((j + 1), -1.2)
        Sigma = np.diag(diagonal)
        # V
        V = np.zeros((self.num_clients, self.dimension))
        samples_per_client = BasicPartitioner().data_imbalance_generator(self.num_clients, int(self.mean_datavol * self.num_clients),self.imbalance)
        # Concept Skew
        if self.alpha>=0:
            Us = np.random.normal(0, self.alpha, self.num_clients)
            W = [np.random.normal(Us[k], 1, (self.dimension, self.num_classes)) for k in range(self.num_clients)]
            b = [np.random.normal(Us[k], 1, self.num_classes) for k in range(self.num_clients)]
        else:
            W = [W_global for _ in range(self.num_clients)]
            b = [b_global for _ in range(self.num_clients)]
        # Feature Skew
        if self.beta>=0:
            B = np.random.normal(0, self.beta, self.num_clients)
            for cid in range(self.num_clients): V[cid] = np.random.normal(B[cid], 1, self.dimension)
        else:
            for cid in range(self.num_clients): V[cid] = v_global
        X_split = [[] for _ in range(self.num_clients)]
        y_split = [[] for _ in range(self.num_clients)]
        optimal_local = [np.concatenate((wk, bk.reshape(1, bk.shape[0])), axis=0).tolist() for wk, bk in zip(W, b)]
        test_data = {'x': [], 'y':[]}
        local_datas = []
        for cid in range(self.num_clients):
            # X_ki~N(v_k, Sigma)
            X_k = np.random.multivariate_normal(V[cid], Sigma, samples_per_client[cid])
            Y_k = np.zeros(samples_per_client[cid], dtype=int)
            for i in range(samples_per_client[cid]):
                # Y_ki = argmax(softmax(W_k x_ki + b_k))
                tmp = np.dot(X_k[i], W[cid]) + b[cid]
                Y_k[i] = np.argmax(np.exp(tmp)/np.sum(np.exp(tmp)))
            X_split[cid] = X_k.tolist()
            y_split[cid] = Y_k.tolist()
            # split dataset to train and testing part
            k = int(samples_per_client[cid]*0.85)
            x_train = X_split[cid][:k]
            y_train = y_split[cid][:k]
            x_test = X_split[cid][k:]
            y_test = y_split[cid][k:]
            test_data['x'].extend(x_test)
            test_data['y'].extend(y_test)
            local_datas.append({'x': x_train, 'y':y_train})
        self.test_data = test_data
        self.local_datas = local_datas
        return X_split, y_split, optimal_local

    def get_task_name(self):
        partitioner_name = 'iid' if self.alpha<0 and self.beta<0 else 'a{:.1f}b{:.1f}'.format(self.alpha, self.beta)
        if self.imbalance>0: partitioner_name += '_imb{:.1f}'.format(self.imbalance)
        return '_'.join(['B-'+self.benchmark, 'P-'+partitioner_name, 'N-'+str(self.num_clients)])



r"""
Synthetic dataset is from the article 'Federated Optimization in Heterogeneous Networks' (http://arxiv.org/abs/1812.06127) of Li et. al.
The para of this benchmark is
Args:::
    alpha (float): Synthetic(alpha, beta), default value is 0
    beta (float): Synthetic(alpha, beta), default value is 0
    num_clients (int): the number of all the clients, default value is 30
    imbalance (float): the degree of data-size-imbalance ($imbalance \in [0,1]$), default value is 0.0
    mean_datavol (int): the mean data size of all the clients, default value is 400
    dimension (int): the dimension of the feature, default value is 60
    num_classes (int): the number of classes in the label set, default value is 10

Example::
```python
    >>> import flgo
    >>> import flgo.benchmark.synthetic_regression as synthetic
    >>> config = {'benchmark':{'name':synthetic, 'para':{'alpha':1., 'beta':1., 'num_clients':30}}}
    >>> flgo.gen_task(config, './my_synthetic11')
```
"""
from flgo.benchmark.synthetic_regression.model import lr

default_model = lr


from flgo.experiment.logger import BasicLogger

class VerticalLogger(BasicLogger):
    r"""Vertical Logger. Vertical FL should use this Logger"""
    def initialize(self):
        """This method is used to record the stastic variables that won't change across rounds (e.g. local_movielens_recommendation data size)"""
        for c in self.participants:
            self.output['client_datavol'].append(len(c.train_data))

    """This logger only records metrics on validation dataset"""
    def log_once(self, *args, **kwargs):
        self.info('Current_time:{}'.format(self.clock.current_time))
        self.output['time'].append(self.clock.current_time)
        test_metric = self.coordinator.test('test')
        val_metric = self.coordinator.test('val')
        for met_name, met_val in test_metric.items():
            self.output['test_' + met_name].append(met_val)
        for met_name, met_test in val_metric.items():
            self.output['val_' + met_name].append(met_val)
        self.show_current_output()


"""
This is a non-official implementation of 'Ditto: Fair and Robust Federated Learning Through Personalization' (https://arxiv.org/abs/2012.04221)
"""
from flgo.algorithm.fedbase import BasicServer, BasicClient
import copy
import torch
from flgo.utils import fmodule

class Server(BasicServer):
    def initialize(self, *args, **kwargs):
        self.init_algo_para({'mu':1.0})
        self.sample_option = 'md'
        self.aggregation_option = 'uniform'

class Client(BasicClient):
    def initialize(self, *args, **kwargs):
        self.model = None

    @fmodule.with_multi_gpus
    def train(self, model):
        model.train()
        # global parameters
        src_model = copy.deepcopy(model)
        src_model.freeze_grad()
        if self.model is None: self.model = copy.deepcopy(model)
        self.model = self.model.to(self.device)
        self.model.train()
        # global solver
        optimizer_global = self.calculator.get_optimizer(model, lr=self.learning_rate, weight_decay=self.weight_decay, momentum=self.momentum)
        optimizer_local = self.calculator.get_optimizer(self.model, lr = self.learning_rate, weight_decay=self.weight_decay, momentum=self.momentum)
        for iter in range(self.num_steps):
            batch_data = self.get_batch_data()
            # global solver
            model.zero_grad()
            loss_global = self.calculator.compute_loss(model, batch_data)['loss']
            loss_global.backward()
            if self.clip_grad>0:torch.nn.utils.clip_grad_norm_(parameters=model.parameters(), max_norm=self.clip_grad)
            optimizer_global.step()
            # local solver
            self.model.zero_grad()
            loss_local = self.calculator.compute_loss(self.model, batch_data)['loss']
            loss_proximal = 0
            for pm, ps in zip(self.model.parameters(), src_model.parameters()):
                loss_proximal += torch.sum(torch.pow(pm - ps, 2))
            loss_local = loss_local + 0.5 * self.mu * loss_proximal
            loss_local.backward()
            if self.clip_grad>0:torch.nn.utils.clip_grad_norm_(parameters=self.model.parameters(), max_norm=self.clip_grad)
            optimizer_local.step()
        self.model = self.model.to('cpu')
        return


from torch import nn
import torch.nn.functional as F
from flgo.utils.fmodule import FModule

class Model(FModule):
    def __init__(self):
        super().__init__()
        self.encoder = nn.Sequential(
            nn.Conv2d(in_channels=1, out_channels=32, kernel_size=5, padding=2),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Conv2d(in_channels=32, out_channels=64, kernel_size=5, padding=2),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Flatten(1),
            nn.Linear(3136, 512),
            nn.ReLU(),
        )
        self.head = nn.Linear(512, 10)

    def forward(self, x):
        x = self.encoder(x)
        x = self.head(x)
        return x

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


"""
train_data (torch.utils.data.Dataset),
test_data (torch.utils.data.Dataset),
and the model (torch.nn.Module) should be implemented here.

"""
import torch.nn

language_pair = (None, None)
tokenizers = {}
vocabs = {}

train_data = None
val_data = None
test_data = None

def get_model(*args, **kwargs) -> torch.nn.Module:
    raise NotImplementedError


from typing import *
from collections.abc import Callable
from torch.utils.data import random_split, Subset
from torchtext.data.functional import to_map_style_dataset
from flgo.benchmark.base import BasicTaskCalculator, BasicTaskGenerator, BasicTaskPipe
import os
try:
    import ujson as json
except:
    import json
import torch

class GeneralCalculator(BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super(GeneralCalculator, self).__init__(device, optimizer_name)
        self.DataLoader = torch.utils.data.DataLoader

    def criterion(self, outputs, targets, ignore_index=-100):
        loss_func = torch.nn.CrossEntropyLoss(ignore_index=ignore_index)
        return loss_func(outputs[1:].view(-1, outputs.shape[-1]), targets[1:].view(-1))

    def compute_loss(self, model, data):
        """
        Args:
            model: the model to train
            data: the training dataset
        Returns: dict of train-one-step's result, which should at least contains the key 'loss'
        """
        sources, targets = self.to_device(data)
        outputs = model(sources, targets)
        loss = self.criterion(outputs, targets, model.ignore_index if hasattr(model, 'ignore_index') else -100)
        return {'loss': loss}

    @torch.no_grad()
    def test(self, model, dataset, batch_size=64, num_workers=0, pin_memory=False):
        """
        Metric = [mean_accuracy, mean_loss]

        Args:
            model:
            dataset:
            batch_size:
        Returns: [mean_accuracy, mean_loss]
        """
        model.eval()
        if batch_size==-1:batch_size=len(dataset)
        data_loader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers, pin_memory=pin_memory)
        total_loss = 0.0
        for batch_id, batch_data in enumerate(data_loader):
            batch_data = self.to_device(batch_data)
            outputs = model(batch_data[0], batch_data[1])
            batch_mean_loss = self.criterion(outputs, batch_data[1], model.ignore_index if hasattr(model, 'ignore_index') else -100).item()
            total_loss += batch_mean_loss * len(batch_data[-1])
        return {'loss':total_loss/len(dataset)}

    def to_device(self, data):
        return data[0].to(self.device), data[1].to(self.device)

    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0, pin_memory=False, drop_last=False, *args, **kwargs):
        if self.DataLoader == None:
            raise NotImplementedError("DataLoader Not Found.")
        return self.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, pin_memory=pin_memory, drop_last=drop_last, collate_fn=self.collate_fn)



import os
import shutil
import urllib.request
import zipfile
from typing import Any
import math
import json
import numpy as np
from torchvision.datasets import MNIST
from PIL import Image
import torch
import torchvision
import torchvision.datasets.utils as tdu
from tqdm import tqdm

transform = torchvision.transforms.Compose([torchvision.transforms.ToTensor(), torchvision.transforms.Normalize((0.9637,), (0.1591,))])
root = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'RAW_DATA', 'MNIST')

def download_from_url(url= None, filepath = '.'):
    """Download dataset from url to filepath."""
    if url:urllib.request.urlretrieve(url, filepath)
    return filepath

def extract_from_zip(src_path, target_path):
    """Unzip the .zip file (src_path) to target_path"""
    f = zipfile.ZipFile(src_path)
    f.extractall(target_path)
    targets = f.namelist()
    f.close()
    return [os.path.join(target_path, tar) for tar in targets]

class FEMNIST(MNIST):
    """
    This dataset is derived from the Leaf repository
    (https://github.com/TalwalkarLab/leaf) pre-processing of the Extended MNIST
    dataset, grouping examples by writer. Details about Leaf were published in
    "LEAF: A Benchmark for Federated Settings" https://arxiv.org/abs/1812.01097.
    """

    def __init__(self, root:str, train:bool=True, transform:Any=None, target_transform:Any=None,
                 download:bool=False):
        super(MNIST, self).__init__(root, transform=transform,
                                    target_transform=target_transform)
        self.train = train
        if self.train:
            data_file = self.training_file
        else:
            data_file = self.test_file
        if not os.path.exists(os.path.join(self.processed_folder, data_file)):
            self.download_and_process()
        self.data, self.targets, self.user_index = torch.load(os.path.join(self.processed_folder, data_file))
        self.id = self.user_index.tolist()

    def __getitem__(self, index):
        img, target = self.data[index], int(self.targets[index])
        img = Image.fromarray(img.numpy(), mode='F')
        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return img, target

    def download_and_process(self):
        """
        Download the raw data and process it and save it in Torch format
        Modified from https://github.com/alibaba/FederatedScope/blob/master/federatedscope/cv/dataset/leaf_cv.py
        """
        os.makedirs(self.raw_folder, exist_ok=True)
        os.makedirs(self.processed_folder, exist_ok=True)
        if not os.path.exists(os.path.join(self.raw_folder, 'all_data')):
            """Download the FEMNIST data if it doesn't exist in processed_folder already."""
            # Download to `self.raw_dir`.
            url = 'https://federatedscope.oss-cn-beijing.aliyuncs.com'
            name = 'femnist_all_data.zip'
            tdu.download_and_extract_archive(f'{url}/{name}', self.raw_folder, remove_finished=True)
            # src_path = download_from_url(f'{url}/{name}', os.path.join(, 'femnist_all_data.zip'))
            # tar_paths = extract_from_zip(src_path, self.raw_folder)

            # with open(os.path.join(self.raw_folder, 'all_data.json'), 'w') as f:
            #     json.dump(all_data, f)
            # os.remove(src_path)
            # shutil.rmtree(tar_paths[0])
        # else:
        #     if not os.path.exists(os.path.join(self.raw_folder, 'all_data')):
        #         tar_paths = extract_from_zip(os.path.join(self.raw_folder, 'femnist_all_data.zip'), self.raw_folder)
        #     else:
        tar_path0 = os.path.join(self.raw_folder, 'all_data')
        tar_paths = [tar_path0] + [os.path.join(tar_path0, r) for r in os.listdir(tar_path0)]
            # with open(os.path.join(self.raw_folder, 'all_data.json'), 'r') as f:
            #     all_data = json.load(f)
        all_data = {
            'users': [],
            'num_samples': [],
            'user_data': {}
        }
        for i in tqdm(range(1, len(tar_paths)), desc='Loading data'):
            with open(tar_paths[i], 'r') as f:
                raw_data = json.load(f)
                all_data['users'].extend(raw_data['users'])
                all_data['num_samples'].extend(raw_data['num_samples'])
                all_data['user_data'].update(raw_data['user_data'])
        """Process Data"""
        Xs = []
        Ys = []
        sample_ids = []
        for idx, (writer, v) in enumerate(all_data['user_data'].items()):
            data, targets = v['x'], v['y']
            Xs.extend(data)
            Ys.extend(targets)
            sample_ids.extend([idx] * len(data))
        Xs = torch.tensor(np.stack(Xs))/255.0
        Ys = torch.LongTensor(np.stack(Ys))
        sample_ids = torch.tensor(np.stack(sample_ids))
        num_samples = sample_ids.shape[0]
        s1 = int(num_samples * 0.9)
        s2 = num_samples - s1
        train_ids, test_ids = torch.utils.data.random_split(sample_ids, [s1, s2])
        train_indices = train_ids.indices
        test_indices = test_ids.indices
        train_data, train_targets, train_sample_id = Xs[train_indices], Ys[train_indices], sample_ids[train_indices]
        test_data, test_targets, test_sample_id = Xs[test_indices], Ys[test_indices], sample_ids[test_indices]
        torch.save((train_data, train_targets, train_sample_id), os.path.join(self.processed_folder, "training.pt"))
        torch.save((test_data, test_targets, test_sample_id), os.path.join(self.processed_folder, "test.pt"))

train_data = FEMNIST(root=root, download=True, train=True, transform=transform)
test_data = FEMNIST(root=root, download=True, train=False, transform=transform)


import torchvision
import os

transform = torchvision.transforms.Compose(
    [torchvision.transforms.ToTensor(),
     torchvision.transforms.Normalize((0.4467, 0.4398, 0.4066), (0.2603, 0.2584, 0.2712))]
)

path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'RAW_DATA', 'STL10')
train_data = torchvision.datasets.STL10(root=path, train=True, download=True, transform=transform)
test_data = torchvision.datasets.STL10(root=path, train=False, download=True, transform=transform)